/* Timer.m
 * Written By:  Thomas Burkholder
 *
 * You may freely copy, distribute, and reuse the code in this example.
 * NeXT disclaims any warranty of any kind, expressed or  implied, as to its
 * fitness for any particular use.
 */

// extended by wave for WavesWorld use

#import "WWTTTimer.h"
#import "WWTCLInterp.h"

// for IB protocols
#import <apps/InterfaceBuilder.h>

#define DEFAULTPERIOD 1.0
#define LOWPERIOD 0.05
#define HIGHPERIOD MAXDOUBLE

#define DEFAULTSYNC 60
#define LOWSYNC 1
#define HIGHSYNC MAXINT

#define DEFAULTSTART 0.0
#define LOWSTART MINDOUBLE
#define HIGHSTART MAXDOUBLE

#define LOWWRAP MINDOUBLE
#define HIGHWRAP MAXDOUBLE

#define DEFAULTINCREMENT 1.0
#define LOWINCREMENT MINFLOAT
#define HIGHINCREMENT MAXFLOAT

#define DEFAULTPRIORITY NX_BASETHRESHOLD
#define LOWPRIORITY 0
#define HIGHPRIORITY 30


//#define MAX(_x,_y) (_x>_y)?_x:_y
//#define MIN(__x,__y) (__x<__y)?__x:__y
#define CONSTRAIN(__a,__low,__high) MIN(MAX(__a,__low),__high)

@implementation WWTTTimer

// function that does what's necessary to the object.  To maximise speed, 
// methodFor: is used.
static void handle(DPSTimedEntry tag, double now, char *userdata)
{
   WWTTTimer *self = (id)userdata;

   self->value += self->incrementBy;
   self->syncCounter++;
   if (self->bFlags.sync && (self->syncCounter >= self->syncValue)) 
   {  [self synchronize:self];
      self->syncCounter = 0;
   }
   if (self->finiteLife)
   {  if (self->conditionalLifespan)
      {  // have our interp evaluate our "reasonForLiving" conditional
	 if (![self checkConditional])  // my reason to live is gone! 
	 {  [self stop:self];
 	 }
         else
         {  if (self->actionFunc)
            {  self->actionFunc(self->target, self->action, self);
            }
            [self updateTclCommand];
            [self updateInterp];
 	 }
      }
      else
      {  if (self->value >= self->wrapValue)
         {  [self stop:self];
         }
         else
         {  if (self->actionFunc)
            {  self->actionFunc(self->target, self->action, self);
            }
            [self updateTclCommand];
            [self updateInterp];
 	 }
      }
   }
   else
   {  if (self->bFlags.wrap && (self->value >= self->wrapValue))
      {  self->value = self->startValue;
      }
      if (self->actionFunc)
      {  self->actionFunc(self->target, self->action, self);
      }
      [self updateTclCommand];
      [self updateInterp];
   }
   
   return;
}

// initializing and freeing
+ initialize  {	if (self == [WWTTTimer class]) { [self setVersion:5];}	return self; }

- init
{
	[super init];

	entry = NULL;
	period = DEFAULTPERIOD;
	priority = DEFAULTPRIORITY;

	bFlags.visibleDebug = NO;
	bFlags.wrap = YES;
	bFlags.sync = NO;

	syncValue = DEFAULTSYNC;
	value = 0.0;
	startValue = DEFAULTSTART;
	wrapValue = 60.0;
	incrementBy = DEFAULTINCREMENT;

	target = nil;
	action = (SEL)0;
	actionFunc = NULL;
	autoStart = NO;
	awakeInIB = NO;
        paused = NO;

  preControlStringSize = 32;
  preControlString = (char *)NXZoneCalloc([self zone], preControlStringSize, sizeof(char));

  controlStringSize = 32;
  controlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
  sprintfControlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));

  postControlStringSize = 32;
  postControlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));

  tclCommandSize = 32;
  tclCommand = (char *)NXZoneCalloc([self zone], tclCommandSize, sizeof(char));
  interp = nil;
  tclCommandDirty = YES;

  finiteLife = NO;

  return self;
}

- awake
{
  [super awake];
  tclCommandDirty = YES;
  paused = NO;
  return self;
}

- free
{
  if (preControlString) { NXZoneFree([self zone], preControlString); }
  if (controlString) { NXZoneFree([self zone], controlString); }
  if (postControlString) { NXZoneFree([self zone], postControlString); }
  if (sprintfControlString) { NXZoneFree([self zone], sprintfControlString); }
  if (tclCommand) { NXZoneFree([self zone], tclCommand); }
  if (entry) 
  {  [self stop:self];
  }
  return [super free];
}

- awakeFromNib
{
   if (autoStart)
   {  [self start:nil];
   }

   // need to add myself to a list inside the WWTCLInterp so that when
   // it is being free'ed it frees us first
   if (interp)
   {  [interp addToManipulatorList:self];
   }

   return self;
}


// oh, the shame...
- finishUnarchiving
{
   if (awakeInIB)
   {  return self;
   }
   awakeInIB = YES;
   if (autoStart)
   {  if ([NXApp conformsTo:@protocol(IB)])
      {  [self perform:@selector(start:) with:nil afterDelay:0.01 cancelPrevious:NO];
      }
   }

   return self;
}


- (BOOL)finiteLife { return finiteLife; }
- setFiniteLife:(BOOL)n { finiteLife = n; return self; }

- (BOOL)conditionalLifespan { return conditionalLifespan; }
- setConditionalLifespan:(BOOL)n { conditionalLifespan = n; return self; }
- (BOOL)checkConditional  {  return [interp exprBoolean:conditional];  }


- reinitializeControlStringFromZone:(NXZone *)zone
{
  preControlString = (char *)NXZoneCalloc(zone, preControlStringSize, sizeof(char));
  controlString = (char *)NXZoneCalloc(zone, controlStringSize, sizeof(char));
  sprintfControlString = (char *)NXZoneCalloc(zone, controlStringSize, sizeof(char));
  postControlString = (char *)NXZoneCalloc(zone, postControlStringSize, sizeof(char));
  tclCommand = (char *)NXZoneCalloc(zone, tclCommandSize, sizeof(char));
  return self;
}

- copyFromZone:(NXZone *)zone
{
  id newCopy = [super copyFromZone:zone];


  [newCopy reinitializeControlStringFromZone:zone];
  [newCopy setPreControlString:preControlString];
  [newCopy setConditional:conditional];
  [newCopy setControlString:controlString];
  [newCopy setPostControlString:postControlString];
  [newCopy setTclCommand:tclCommand];

  return newCopy;
}

- setInterp:newInterp { interp = newInterp; return self; }
- interp { return interp; }

- resizeTclCommandForArgVector:(char *)argOrderVector
{
  BOOL    done = NO;
  

  if (!done && !strcmp("d", argOrderVector))
  {  while (tclCommandSize < (controlStringSize + 32))
     {  if (!tclCommandSize)
	{  tclCommandSize = 32;
           tclCommand = (char *)NXZoneCalloc([self zone], tclCommandSize, sizeof(char));
        }
        else
        {  tclCommandSize *= 2;
           tclCommand = (char *)NXZoneRealloc([self zone], tclCommand, tclCommandSize);
        }
     }
     done = YES;
  }
  if (!done && !strcmp("f", argOrderVector))
  {  while (tclCommandSize < (controlStringSize + 32) )
     {  if (!tclCommandSize)
	{  tclCommandSize = 32;
           tclCommand = (char *)NXZoneCalloc([self zone], tclCommandSize, sizeof(char));
        }
        else
        {  tclCommandSize *= 2;
           tclCommand = (char *)NXZoneRealloc([self zone], tclCommand, tclCommandSize);
        }
     }
     done = YES;
  }

  return self;
}


- updateTclCommand
{
  char    *controlStringPtr, *sprintfControlStringPtr;
  BOOL    done;
  int     argOrderSize = 32;
  char    *argOrderVector = (char *)NXZoneCalloc([self zone], argOrderSize, sizeof(char));
  int     argOrderOffset = 0;


  if (!controlString)
  {  return nil;
  }

  // we want to grovel over the controlString, 
  // for right now, we really only tend to have the following patterns 
  // in the controlString: 1 %f, 1 %d
  controlStringPtr = controlString;
  sprintfControlStringPtr = sprintfControlString;
  strcpy(sprintfControlString, controlString);
  while (*controlStringPtr)
  {  if (*controlStringPtr == '%')
     {  sprintfControlStringPtr++;  controlStringPtr++;
        if (*controlStringPtr)
        {  switch (*controlStringPtr)
           {  case 'd':	while (argOrderOffset >= argOrderSize)
                        {  argOrderSize *= 2;
                           argOrderVector = (char *)NXZoneRealloc([self zone], argOrderVector, argOrderSize);
                        }
                        argOrderVector[argOrderOffset++] = 'd';
                        sprintfControlStringPtr++; controlStringPtr++;
			break;
              case 'f':	while (argOrderOffset >= argOrderSize)
                        {  argOrderSize *= 2;
                           argOrderVector = (char *)NXZoneRealloc([self zone], argOrderVector, argOrderSize);
                        }
                        argOrderVector[argOrderOffset++] = 'f';
                        sprintfControlStringPtr++; controlStringPtr++;
			break;
              case '%':	sprintfControlStringPtr++; controlStringPtr++;
                        break;
               default:	NXLogError("WWTTTimer: unknown specifier <%c> in controlString.\n", *controlStringPtr);
                        sprintfControlStringPtr++; controlStringPtr++;
			break;
           }
       }
     }
     else
     {  sprintfControlStringPtr++;  controlStringPtr++;
     }
  }
  argOrderVector[argOrderOffset] = 0;

  done = NO;

  if (!done && !strcmp("d", argOrderVector))
  {  [self resizeTclCommandForArgVector:argOrderVector];
     sprintf(tclCommand, sprintfControlString, [self intValue]);
     done = YES;
  }
  if (!done && !strcmp("f", argOrderVector))
  {  [self resizeTclCommandForArgVector:argOrderVector];
     sprintf(tclCommand, sprintfControlString, [self floatValue]);
     done = YES;
  }

  if (!done)
  {  while (tclCommandSize < controlStringSize)
     {  if (!tclCommandSize)
	{  tclCommandSize = 32;
           tclCommand = (char *)NXZoneCalloc([self zone], tclCommandSize, sizeof(char));
        }
        else
        {  tclCommandSize *= 2;
           tclCommand = (char *)NXZoneRealloc([self zone], tclCommand, tclCommandSize);
        }
     }
     strcpy(tclCommand, controlString);
  }
  tclCommandDirty = NO;

  NXZoneFree([self zone], argOrderVector);

  return self;
}


- updateInterp
{
  if ([interp eval:[self tclCommand]])
  {  // actually, if this fails, we should stop the timer, right?
     NXLogError("WWTTTimer failed to updateInterp.  Stopping...");
     [self stop:nil];
     return nil;
  }
  return self;
}


- setPreControlString:(const char *)str 
{ 
  int   cnt;


  if (!str) { return self; }
  if (preControlStringSize <= 0) 
  {  preControlStringSize = 32; 
     preControlString = (char *)NXZoneCalloc([self zone], preControlStringSize, sizeof(char));
  }
  cnt = strlen(str);
  while (cnt >= preControlStringSize)
  {  preControlStringSize *= 2;
     preControlString = (char *)NXZoneRealloc([self zone], preControlString, preControlStringSize);
  }
  strcpy(preControlString, str); 
  return self; 
}
//
- (const char *)preControlString { return preControlString; }

- setConditional:(const char *)str 
{ 
  int   cnt;


  if (!str) { return self; }
  if (conditionalSize <= 0) 
  {  conditionalSize = 32; 
     conditional = (char *)NXZoneCalloc([self zone], conditionalSize, sizeof(char));
  }
  cnt = strlen(str);
  while (cnt >= conditionalSize)
  {  conditionalSize *= 2;
     conditional = (char *)NXZoneRealloc([self zone], conditional, conditionalSize);
  }
  strcpy(conditional, str); 
  return self; 
}
//
- (const char *)conditional { return conditional; }

- setControlString:(const char *)str 
{ 
  int   cnt;


  if (!str) { return self; }
  if (controlStringSize <= 0) 
  {  controlStringSize = 32; 
     controlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
  }
  cnt = strlen(str);
  while (cnt >= controlStringSize)
  {  controlStringSize *= 2;
     controlString = (char *)NXZoneRealloc([self zone], controlString, controlStringSize);
  }
  if (sprintfControlString)
  {  sprintfControlString = (char *)NXZoneRealloc([self zone], sprintfControlString, controlStringSize);
  }
  else
  {  sprintfControlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
  }
  strcpy(controlString, str); 
  tclCommandDirty = YES;
  return self; 
}
//
- (const char *)controlString { return controlString; }

- setPostControlString:(const char *)str 
{ 
  int   cnt;


  if (!str) { return self; }
  if (postControlStringSize <= 0) 
  {  postControlStringSize = 32; 
     postControlString = (char *)NXZoneCalloc([self zone], postControlStringSize, sizeof(char));
  }
  cnt = strlen(str);
  while (cnt >= postControlStringSize)
  {  postControlStringSize *= 2;
     postControlString = (char *)NXZoneRealloc([self zone], postControlString, postControlStringSize);
  }
  strcpy(postControlString, str); 
  return self; 
}
//
- (const char *)postControlString { return postControlString; }

- setTclCommand:(const char *)str 
{ 
  int   cnt;


  if (!str) { return self; }
  if (tclCommandSize <= 0) 
  {  tclCommandSize = 32; 
     tclCommand = (char *)NXZoneCalloc([self zone], tclCommandSize, sizeof(char));
  }
  cnt = strlen(str);
  while (cnt >= tclCommandSize)
  {  tclCommandSize *= 2;
     tclCommand = (char *)NXZoneRealloc([self zone], tclCommand, tclCommandSize);
  }
  strcpy(tclCommand, str); 
  return self; 
}
//
- (const char *)tclCommand 
{ 
  if (tclCommandDirty)
  {  [self updateTclCommand];
  }
  return tclCommand; 
}


// debugging methods
- logError:(char *)str
{
	if (bFlags.visibleDebug)
		NXRunAlertPanel("ERROR",str,NULL,NULL,NULL);
	NXLogError(str);
	return self;
}

// private methods
- stopEntry
{
	if (entry == NULL) {
		[self logError:"Stop called on stopped timer!\n"];
		return nil;
	} else {
		DPSRemoveTimedEntry(entry);
		entry = NULL;
	}
	return self;
}

- startEntry
{
	if (entry != NULL) {
		[self stopEntry];
	}

	// get function for method, so we can be super-efficient
	if (target && action)
		actionFunc = [target methodFor:action];

	// make a new timed entry
	entry = DPSAddTimedEntry(period,(DPSTimedEntryProc)handle,self,priority);

	// check it
	if ((int)entry == -1) {
		[self logError:"Couldn't start timer!\n"];
		entry = NULL;
		return nil;
	}
	return self;
}

// actions
- start:sender
{
   if (startButton) {  [startButton setIntValue:1]; } 
   if (stopButton) {  [stopButton setIntValue:0]; } 
   if (pauseButton) {  [pauseButton setIntValue:0]; } 

   if (paused)
   {  return [self resume:sender];
   }

   // stop timer if necessary
   value = startValue;
   syncTime = [self currentTime];
   syncCounter = 0;

   // before we actually start, we want to send our preControlString to the interp 
   [interp eval:preControlString];

   [self startEntry];
   return self;
}

- stop:sender
{
   if (startButton) {  [startButton setIntValue:0]; } 
   if (stopButton) {  [stopButton setIntValue:1]; } 
   if (pauseButton) {  [pauseButton setIntValue:0]; } 

   if (paused)
   {  paused = NO;
   }
   [self stopEntry];
   // after we actually stop, we want to send our postControlString to the interp 
   [interp eval:postControlString];
    return self;
}

- pause:sender
{
   if (paused)
   {  paused = NO;
      if (startButton) {  [startButton setIntValue:1]; } 
      if (stopButton) {  [stopButton setIntValue:0]; } 
      if (pauseButton) {  [pauseButton setIntValue:0]; } 
      return [self startEntry];
   }
   if (startButton) {  [startButton setIntValue:0]; } 
   if (stopButton) {  [stopButton setIntValue:0]; } 
   if (pauseButton) {  [pauseButton setIntValue:1]; } 

   paused = YES;
   return [self stopEntry];
}

- resume:sender
{
   if (startButton) {  [startButton setIntValue:1]; } 
   if (stopButton) {  [stopButton setIntValue:0]; } 
   if (pauseButton) {  [pauseButton setIntValue:0]; } 

   if (paused)
   {  paused = NO;
   }
   return [self startEntry];
}

// just in case anyone wants to fool with the timer directly
- (DPSTimedEntry) entry
{
	return entry;
}

- (unsigned int)currentTime
{
	struct timeval curTime;
	gettimeofday (&curTime, NULL);
	return (curTime.tv_sec) * 1000 + curTime.tv_usec / 1000;
}

- synchronize:sender;
{
	int correctedTicks = (([self currentTime]-syncTime)/((int)(period*1000)));
	if (bFlags.wrap) {
		correctedTicks = correctedTicks % (int)((wrapValue-startValue)/incrementBy);
	}
	value = correctedTicks*incrementBy;
	return self;
}

// methods to get/set the value of the timer
- (float)floatValue { return value; }
- setFloatValue:(float)aValue { value = aValue; return self; }
- (int)intValue { return (int)value; }
- setIntValue:(int)aValue { value = (float)aValue; return self; }
- (const char *)stringValue { static char str[20]; sprintf(str,"%f",value); return str; }
- setStringValue:(const char *)aValue { value = atof(aValue); return self; }
- (double)doubleValue { return (double)value; }
- setDoubleValue:(double)aValue { value = (double)aValue; return self; }

// methods to get/set behaviours
- (BOOL)autoStart { return autoStart; }
- setAutoStart:(BOOL)flag { autoStart = flag; return self; }
- setIncrementBy:(float)aValue { incrementBy = aValue; return self; }
- (float)incrementBy { return incrementBy; }
- setStartValue:(float)aValue { startValue = aValue; return self; }
- (float)startValue { return startValue; }
- setWrap:(BOOL)yn { bFlags.wrap = yn; return self; }
- (BOOL)wrap { return (BOOL)(bFlags.wrap); }
- setWrapValue:(float)aValue { wrapValue = CONSTRAIN(aValue,LOWWRAP,HIGHWRAP); return self; }
- (float)wrapValue { return wrapValue; }
- (int)syncValue { return syncValue; }
- setSyncValue:(int)aValue { syncValue = CONSTRAIN(aValue,LOWSYNC,HIGHSYNC); return self; }
- (BOOL)sync { return (BOOL)(bFlags.sync); }
- setSync:(BOOL)yn { bFlags.sync = yn; return self; }
- (BOOL)visibleDebug { return (BOOL)(bFlags.visibleDebug); }
- setVisibleDebug:(BOOL)yn { bFlags.visibleDebug = yn; return self; }
- (int)priority { return priority; }
- setPriority:(int)aPriority { priority = CONSTRAIN(aPriority,LOWPRIORITY,HIGHPRIORITY); return self; }
- (double)period { return period; }
- setPeriod:(double)aPeriod { period = CONSTRAIN(aPeriod,LOWPERIOD,HIGHPERIOD); return self; }
- (SEL)action { return action; }

- setAction:(SEL)anAction
{
	action = anAction;
	// in case someone decides to change targets while the timer is operational
	if (target && action)
		actionFunc = [target methodFor:action];
	return self; 
}

- target
{
	return target;
}

- setTarget:anObject
{
	target = anObject;
	// in case someone decides to change targets while the timer is operational
	if (target && action)
		actionFunc = [target methodFor:action];
	return self;
}

- takePeriod:sender  {  period = [sender doubleValue]; return self; }
- takeStartValue:sender   {  startValue = [sender floatValue]; return self; }
- takeWrapValue:sender   {  wrapValue = [sender floatValue]; return self; }
- takeIncrementBy:sender   {  incrementBy = [sender floatValue]; return self; }
- takeFiniteLife:sender { finiteLife = (BOOL)[sender intValue]; return self; }


// archival methods
- read:(NXTypedStream *)stream
{
   [super read:stream];
   switch (NXTypedStreamClassVersion(stream, "WWTTTimer")) 
   {  case 1:
	  NXReadTypes(stream,
		      "@:diciffffc", 
		      &target, &action, &period,
		      &priority, &bFlags, &syncValue,
		      &startValue, &wrapValue, &value, 
		      &incrementBy, &autoStart);
	  controlStringSize = 32;
	  controlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
	  sprintfControlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
	  tclCommandSize = 32;
	  tclCommand = (char *)NXZoneCalloc([self zone], tclCommandSize, sizeof(char));
	  interp = nil;
          finiteLife = NO;
          conditional = NULL;
          conditionalSize = 0;
          conditionalLifespan = NO;
          break;
      case 2:
	  NXReadTypes(stream,
		      "@:diciffffc", 
		      &target, &action, &period,
		      &priority, &bFlags, &syncValue,
		      &startValue, &wrapValue, &value, 
		      &incrementBy, &autoStart);
	  NXReadType(stream, "*", &controlString);
          if (controlString) 
          {  controlStringSize = strlen(controlString) + 1;
             sprintfControlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
          }
          else
          {  controlStringSize = 0;
             sprintfControlString = NULL;
          }
	  NXReadType(stream, "*", &tclCommand);
          if (tclCommand) 
          {  tclCommandSize = strlen(tclCommand) + 1;
          }
          else
          {  tclCommandSize = 0;
          }
	  interp = NXReadObject(stream);
          finiteLife = NO;
          conditional = NULL;
          conditionalSize = 0;
          conditionalLifespan = NO;
          break;
      case 3:
	  NXReadTypes(stream,
		      "@:diciffffc", 
		      &target, &action, &period,
		      &priority, &bFlags, &syncValue,
		      &startValue, &wrapValue, &value, 
		      &incrementBy, &autoStart);
	  NXReadType(stream, "*", &controlString);
          if (controlString) 
          {  controlStringSize = strlen(controlString) + 1;
             sprintfControlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
          }
          else
          {  controlStringSize = 0;
             sprintfControlString = NULL;
          }
	  NXReadType(stream, "*", &tclCommand);
          if (tclCommand) 
          {  tclCommandSize = strlen(tclCommand) + 1;
          }
          else
          {  tclCommandSize = 0;
          }
	  interp = NXReadObject(stream);
	  NXReadType(stream, "c", &finiteLife);
          conditional = NULL;
          conditionalSize = 0;
          conditionalLifespan = NO;
          break;
      case 4:
	  NXReadTypes(stream,
		      "@:diciffffc", 
		      &target, &action, &period,
		      &priority, &bFlags, &syncValue,
		      &startValue, &wrapValue, &value, 
		      &incrementBy, &autoStart);
	  NXReadType(stream, "*", &controlString);
          if (controlString) 
          {  controlStringSize = strlen(controlString) + 1;
             sprintfControlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
          }
          else
          {  controlStringSize = 0;
             sprintfControlString = NULL;
          }
	  NXReadType(stream, "*", &tclCommand);
          if (tclCommand) 
          {  tclCommandSize = strlen(tclCommand) + 1;
          }
          else
          {  tclCommandSize = 0;
          }
	  interp = NXReadObject(stream);
	  NXReadType(stream, "c", &finiteLife);

	  NXReadType(stream, "*", &preControlString);
          if (preControlString) 
          {  preControlStringSize = strlen(preControlString) + 1;
          }
          else
          {  preControlStringSize = 0;
          }

	  NXReadType(stream, "*", &postControlString);
          if (postControlString) 
          {  postControlStringSize = strlen(postControlString) + 1;
          }
          else
          {  postControlStringSize = 0;
          }
          conditional = NULL;
          conditionalSize = 0;
          conditionalLifespan = NO;
          break;
      default:
	  NXReadTypes(stream,
		      "@:diciffffc", 
		      &target, &action, &period,
		      &priority, &bFlags, &syncValue,
		      &startValue, &wrapValue, &value, 
		      &incrementBy, &autoStart);
	  NXReadType(stream, "*", &controlString);
          if (controlString) 
          {  controlStringSize = strlen(controlString) + 1;
             sprintfControlString = (char *)NXZoneCalloc([self zone], controlStringSize, sizeof(char));
          }
          else
          {  controlStringSize = 0;
             sprintfControlString = NULL;
          }
	  NXReadType(stream, "*", &tclCommand);
          if (tclCommand) 
          {  tclCommandSize = strlen(tclCommand) + 1;
          }
          else
          {  tclCommandSize = 0;
          }
	  interp = NXReadObject(stream);
	  NXReadType(stream, "c", &finiteLife);

	  NXReadType(stream, "*", &preControlString);
          if (preControlString) 
          {  preControlStringSize = strlen(preControlString) + 1;
          }
          else
          {  preControlStringSize = 0;
          }

	  NXReadType(stream, "*", &postControlString);
          if (postControlString) 
          {  postControlStringSize = strlen(postControlString) + 1;
          }
          else
          {  postControlStringSize = 0;
          }

	  NXReadType(stream, "*", &conditional);
          if (conditional) 
          {  conditionalSize = strlen(conditional) + 1;
          }
          else
          {  conditionalSize = 0;
          }

	  NXReadType(stream, "c", &conditionalLifespan);

          break;
   }
   entry = NULL;
   actionFunc = NULL;
   awakeInIB = NO;
   return self;
}

- write:(NXTypedStream *)stream
{
	[super write:stream];
	NXWriteTypes(stream,
		     "@:diciffffc", 
		     &target, &action, &period,
		     &priority, &bFlags, &syncValue,
		     &startValue, &wrapValue, &value,
		     &incrementBy, &autoStart);
	  NXWriteType(stream, "*", &controlString);
	  NXWriteType(stream, "*", &tclCommand);
	  NXWriteObjectReference(stream, interp);
	  NXWriteType(stream, "c", &finiteLife);
	  NXWriteType(stream, "*", &preControlString);
	  NXWriteType(stream, "*", &postControlString);
	  NXWriteType(stream, "*", &conditional);
	  NXWriteType(stream, "c", &conditionalLifespan);
	return self;
}

@end
